// ***********************************************************************************
//   Custom DPA Handler code example - User peripheral implementation - i2c          *
// ***********************************************************************************
//
// File:    $RCSfile: CustomDpaHandler-UserPeripheral-i2c.c,v $
// Version: $Revision: 1.6 $
// Date:    $Date: 2015/08/24 11:11:43 $
//
// Revision history:
//   2015/08/05  Release for DPA 2.20
//
// *********************************************************************

// Online DPA documentation http://www.iqrf.org/DpaTechGuide/

// This example implements the user peripheral reading from MCP9802 at DDC-SE-01
// PNUM = 0x20 and PCMD = 0 returns 2 bytes with result read from MCP9802
// Based on example DDC-SE-01-i2c.c

// Default IQRF include (modify the path according to your setup)
#include "template-basic.h"

// Default DPA header (modify the path according to your setup)
#include "DPA.h"
// Default Custom DPA Handler header (modify the path according to your setup)
#include "DPAcustomHandler.h"

//############################################################################################

void i2c_init();
void i2c_shutdown();
void i2c_waitForIdle();
void i2c_start();
void i2c_repStart();
void i2c_stop();
uns8 i2c_read(bit ack);
void i2c_write(uns8 i2cWriteData);
uns8 check_crc(uns8 data0, uns8 data1, uns8 checksum);
uns16 read_dht(uns8 valueType);

#define PWR_SENSOR_TRIS 	TRISC.7
#define PWR_SENSOR_IO   	LATC.7
#define DHT_ADR         	0b10000000
#define POLYNOMIAL 			0x131  //P(x)=x^8+x^5+x^4+1 = 100110001
#define DHT_TEMPERATURE     0b11100011
#define DHT_HUMIDITY		0b11100101

#define CRC_OK 				0xC0
#define CRC_ERR 			0x80  

//hw identifier
#define HWPIDDEF			0x8001  //DHT sensor
#define HWPIDVER			0x0001  //DHT sensor


// Must be the 1st defined function in the source code in order to be placed at the correct FLASH location!
//############################################################################################
bit CustomDpaHandler()
//############################################################################################
{
	// Handler presence mark
	clrwdt();

	// Detect DPA event to handle
	switch (GetDpaEvent())
	{
		// -------------------------------------------------
		case DpaEvent_Interrupt:
			// Do an extra quick background interrupt work
			// ! The time spent handling this event is critical.If there is no interrupt to handle return immediately otherwise keep the code as fast as possible.
			// ! Make sure the event is the 1st case in the main switch statement at the handler routine.This ensures that the event is handled as the 1st one.
			// ! It is desirable that this event is handled with immediate return even if it is not used by the custom handler because the Interrupt event is raised on every MCU interrupt and the empty return handler ensures the shortest possible interrupt routine response time.
			// ! Only global variables or local ones marked by static keyword can be used to allow reentrancy.
			// ! Make sure race condition does not occur when accessing those variables at other places.
			// ! Make sure( inspect.lst file generated by C compiler ) compiler does not create any hidden temporary local variable( occurs when using division, multiplication or bit shifts ) at the event handler code.The name of such variable is usually Cnumbercnt.
			// ! Do not call any OS functions except getINFx() and setINDFx().
			// ! Do not use any OS variables especially for writing access.
			// ! All above rules apply also to any other function being called from the event handler code, although calling any function from Interrupt event is not recommended because of additional MCU stack usage.

		DpaHandleReturnTRUE:
			return TRUE;

			// -------------------------------------------------
		case DpaEvent_Init:
		case DpaEvent_AfterSleep:
			// Called after woken up after sleep
			i2c_init();
			break;

			// -------------------------------------------------
		case DpaEvent_BeforeSleep:
			// Called before going to sleep
			i2c_shutdown();
			break;

			// -------------------------------------------------
		case DpaEvent_DpaRequest:
			// Called to interpret DPA request for peripherals
			// -------------------------------------------------
			// Peripheral enumeration
			if (IsDpaEnumPeripheralsRequest())
			{
				// We implement 1 user peripheral
				_DpaMessage.EnumPeripheralsAnswer.UserPerNr = 1;
				_DpaMessage.EnumPeripheralsAnswer.HWPID = HWPIDDEF;
				_DpaMessage.EnumPeripheralsAnswer.HWPIDver = HWPIDVER;

				goto DpaHandleReturnTRUE;
			}
			// -------------------------------------------------
			// Get information about peripheral
			else if (IsDpaPeripheralInfoRequest())
			{
				if (_PNUM == PNUM_USER + 0)
				{
					_DpaMessage.PeripheralInfoAnswer.PerT = PERIPHERAL_TYPE_USER_AREA;
					_DpaMessage.PeripheralInfoAnswer.PerTE = PERIPHERAL_TYPE_EXTENDED_READ;
					goto DpaHandleReturnTRUE;
				}

				break;
			}
			// -------------------------------------------------
			else
			{
				// Handle peripheral command
				if (_PNUM == PNUM_USER + 0)
				{
					// Check command
					switch (_PCMD)
					{
						case 0:
							// -------------------------------------------------
							// Read temperature
							if (_DpaDataLength != 0)
								DpaApiReturnPeripheralError(ERROR_DATA_LEN);

							uns16 temperature  @ _DpaMessage.Response.PData;
							
							temperature = read_dht(DHT_TEMPERATURE);

							_DpaDataLength = sizeof(temperature);
							goto DpaHandleReturnTRUE;


						case 1:
							//Read humidity
							if (_DpaDataLength != 0)
								DpaApiReturnPeripheralError(ERROR_DATA_LEN);

							uns16 humidity  @ _DpaMessage.Response.PData;

							humidity = read_dht(DHT_HUMIDITY);

							_DpaDataLength = sizeof(temperature);
							goto DpaHandleReturnTRUE;


						default:
							// -------------------------------------------------
							// Invalid command
							DpaApiReturnPeripheralError(ERROR_PCMD);
					}
				}
			}
	}

	return FALSE;
}

//############################################################################################

#pragma library 1	// Compile only used functions

// *********************************************************************

void writeToSSPCON2(uns8 value @ param4.high8)
{
	writeToRAM(&SSPCON2, value);
}
//----------------------------------------------------------------------

void writeOredToSSPCON2(uns8 value @ param4.high8)
{
	writeToSSPCON2(SSPCON2 | value);
}

//----------------------------------------------------------------------

void i2c_init()
{
	PORTC = 0x80;                       // port

	TRISC.3 = 1;                        // SCL as input (SIM C6)
	TRISC.4 = 1;                        // SDA as input (SIM C7)

	PWR_SENSOR_TRIS = 0;                // sensor power as output (SIM C8)
	TRISC.5 = 1;                        // shared with SIM C8

	TRISA.5 = 1;                        // sensor ALERT as input (SIM C5)
	TRISB.4 = 1;                        // shared with SIM C5
	TRISC.6 = 1;                        // shared with SIM C5

	writeToRAM(&SSPCON1, 0x38);		  // I2C master mode     SSPCON = 0b00111000
	writeToSSPCON2(0x00);

	SSPADD = (F_OSC / (50000L * 4)) - 1;	// 50 kHz SCL frequency

	SMP = 1;                            // Disable slew rate control
}

//------------------------------------------------------------------------

void i2c_shutdown()
{
	writeToRAM(&SSPCON1, 0x00);		  // I2C master mode     SSPCON = 0
}

//------------------------------------------------------------------------
void i2c_waitForIdle()
{
	while (SSPCON2 & 0x1F);           // Wait for idle and not writing
	while (RW_);                      // Wait for idle and not writing
}
//----------------------------------------------------------------------

void i2c_start()
{
	i2c_waitForIdle();
	writeOredToSSPCON2(0x01);		// SEN = 1
}
//----------------------------------------------------------------------

void i2c_repStart()
{
	i2c_waitForIdle();
	writeOredToSSPCON2(0x02);		// RSEN = 1
}
//----------------------------------------------------------------------

void i2c_stop()
{
	i2c_waitForIdle();
	writeOredToSSPCON2(0x04);		// PEN = 1
}
//----------------------------------------------------------------------

uns8 i2c_read(bit ack)
{
	i2c_waitForIdle();
	writeOredToSSPCON2(0x08);		// RCEN = 1

	i2c_waitForIdle();

	uns8 i2cReadData @ userReg0;
	i2cReadData = SSPBUF;

	i2c_waitForIdle();

	if (ack)
		writeToSSPCON2(SSPCON2 & 0xDF); // Acknowledge, ACKDT = 0
	else
		writeOredToSSPCON2(0x20);		  // Not acknowledge, ACKDT = 1 

	writeOredToSSPCON2(0x10);		  // Send acknowledge sequence, ACKEN = 1 
	return i2cReadData;
}
//----------------------------------------------------------------------

void i2c_write(uns8 i2cWriteData @ param2)
{
	i2c_waitForIdle();
	SSPBUF = i2cWriteData;
}
//----------------------------------------------------------------------
uns8 check_crc(uns8 data0, uns8 data1, uns8 checksum)
{
	uns8 crc = 0;
	uns8 i;

	crc ^= data0;
	for (i = 8; i > 0; --i)
	{
		if (crc & 0x80) crc = (crc << 1) ^ POLYNOMIAL;
		else crc = (crc << 1);
	}

	crc ^= data1;
	for (i = 8; i > 0; --i)
	{
		if (crc & 0x80) crc = (crc << 1) ^ POLYNOMIAL;
		else crc = (crc << 1);
	}

	if (crc != checksum)
	{
		return CRC_ERR;
	}
	else
	{
		return CRC_OK;
	}
}
//----------------------------------------------------------------------
uns16 read_dht(uns8 valueType)
{
	i2c_start();
	i2c_write(DHT_ADR);
	i2c_write(valueType);
	i2c_stop();

	i2c_start();
	i2c_write(DHT_ADR | 1);

	uns16 val;
	uns8 crc;
	
	val.high8 = i2c_read(TRUE);
	val.low8 = i2c_read(TRUE);

	crc = i2c_read(FALSE);
	
	i2c_stop();
		
	uns8 crcStatus = check_crc(val.high8, val.low8, crc);
	
	val >>= 2;
	val.high8 |= crcStatus;

	return val;
}

//----------------------------------------------------------------------


// ############################################################################################
#pragma library 0	// Compile all
//############################################################################################
// Default Custom DPA Handler header; 2nd include to implement Code bumper to detect too long code of the Custom DPA Handler (modify the path according to your setup) 
#include "DPAcustomHandler.h"
//############################################################################################
