﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace IqrfDpaLib
{
    public enum SensorCmd
    {
        Temperature = 0,
        Humidity = 1
    }
    
    public struct DigitalIoDirectionSettings
    {
        public byte Port { get; set; } //port .... A = 0x00 C = 0x02   
        public byte Mask { get; set; } //mask .... 1 selected
        public byte Direction { get; set; } //direction .... 0 output
    }

    public struct DigitalIoValueSettings
    {
        public byte Par1 { get; set; } //port or delay(0xFF)
        public byte Par2 { get; set; } //mask or 1st byte of delay
        public byte Par3 { get; set; } //value or 2nd byte of delay
    }

    public struct DigitalIoState
    {
        public byte PortA { get; set; } 
        public byte PortB { get; set; } 
        public byte PortC { get; set; } 
    }

    public struct ConfirmationContent
    {
        public byte DPA_Val { get; set; }
        public byte Hops { get; set; }
        public byte TimeSlotLength { get; set; }
        public byte HopsRsp { get; set; }

        public override string ToString()
        {
            return string.Format("Confirmation Hops = {0} HopsRsp = {1} TimeSlotLength {2}", Hops, HopsRsp, TimeSlotLength);
        }
    }

    public class DpaFrame
    {
        public static int SensorReq(byte[] buffer, byte address, SensorCmd cmd)
        {
            buffer[0] = address;
            buffer[1] = 0;

            buffer[2] = 0x20;
            buffer[3] = (byte) cmd;

            buffer[4] = 0xFF;
            buffer[5] = 0xFF;

            return 6;
        }

        public static float ParseSensorRsp(byte[] buffer, SensorCmd cmd)
        {
            if (buffer[3] != (byte)((byte)cmd | 0x80) || buffer[6] != 0)
            {
                throw new InvalidDataException("Error in response structure.");
            }

            ushort rawValue = (ushort)(buffer[8] | buffer[9] << 8);
            
            if ((rawValue & 0xC000) != 0xC000)
            {
                throw new InvalidDataException("I2C communication error on sensor.");
            }

            rawValue ^= 0xC000;

            float outputValue = 0;

            switch (cmd)
            {
                case SensorCmd.Humidity:
                    outputValue = (float)(-6.0 + 125.0 / 16384.0 * rawValue);
                    break;

                case SensorCmd.Temperature:
                    outputValue = (float)(-46.85 + 175.72 / 16384.0 * rawValue);
                    break;
                     
            }

            return outputValue;
        }

        public static int IoDirectionReq(byte[] buffer, byte address, DigitalIoDirectionSettings[] directionSettings)
        {
            buffer[0] = address;
            buffer[1] = 0;

            buffer[2] = 0x09;
            buffer[3] = 0x00;

            buffer[4] = 0xFF;
            buffer[5] = 0xFF;

            for (int i = 0; i < directionSettings.Length; i++)
            {
                buffer[3 * i + 6] = directionSettings[i].Port; 
                buffer[3 * i + 7] = directionSettings[i].Mask; 
                buffer[3 * i + 8] = directionSettings[i].Direction;
            }

            return 6 + 3 * directionSettings.Length;
        }

        public static void ValidateIoDirectionRsp(byte[] buffer)
        {
            if (buffer[3] != 0x80 || buffer[6] != 0)
            {
                throw new InvalidDataException("Error in response structure. I/O driection.");
            }
        }

        public static int IoSetReq(byte[] buffer, byte address, DigitalIoValueSettings[] valueSettings)
        {
            buffer[0] = address;
            buffer[1] = 0;

            buffer[2] = 0x09;
            buffer[3] = 0x01;

            buffer[4] = 0xFF;
            buffer[5] = 0xFF;

            for (int i = 0; i < valueSettings.Length; i++)
            {
                buffer[3 * i + 6] = valueSettings[i].Par1;
                buffer[3 * i + 7] = valueSettings[i].Par2;
                buffer[3 * i + 8] = valueSettings[i].Par3;
            }

            return 6 + 3 * valueSettings.Length;
        }

        public static void ValidateIoSetRsp(byte[] buffer)
        {
            if (buffer[3] != 0x81 || buffer[6] != 0)
            {
                throw new InvalidDataException("Error in response frame structure. I/O get.");
            }
        }

        public static int IoGetReq(byte[] buffer, byte address)
        {
            buffer[0] = address;
            buffer[1] = 0;

            buffer[2] = 0x09;
            buffer[3] = 0x02;

            buffer[4] = 0xFF;
            buffer[5] = 0xFF;

            return 6;
        }

        public static DigitalIoState ParseIoGetReq(byte[] buffer)
        {
            if (buffer[3] != 0x82 || buffer[6] != 0)
            {
                throw new InvalidDataException("Error in response frame structure. I/O get.");
            }

            var state = new DigitalIoState();

            state.PortA = buffer[8];
            state.PortB = buffer[9];
            state.PortC = buffer[10];
            
            return state;
        }

        public static ConfirmationContent ParseConfirmationRsp(byte[] buffer)
        {
            if (buffer[6] != 0xFF)
            {
                throw new InvalidDataException("Error in response frame structure. I/O get.");
            }

            return new ConfirmationContent { DPA_Val = buffer[7], Hops = buffer[8], TimeSlotLength = buffer[9], HopsRsp = buffer[10] };
        }
    }
}
