﻿using System;
using System.IO;
using System.IO.Ports;
using System.Threading;

namespace IqrfDpaLib
{
    public class IqrfUartComm
    {
        private SerialPort m_SerialPort;
        private byte[] m_RxBuffer = new byte[1024];
        private byte[] m_TxBuffer = new byte[1024];
        private int m_Offset;
        private TimeSpan m_ReadTimeout;

        public IqrfUartComm(string portName, int baudRate, TimeSpan readTimeout)
        {
            m_SerialPort = new SerialPort(portName, baudRate);
            m_ReadTimeout = readTimeout;
            m_Offset = 0;
        }

        private void Open()
        {
            if (!m_SerialPort.IsOpen)
            {
                for (int i = 0; i < 3; i++)
                {
                    try
                    {
                        m_SerialPort.Open();
                        break;
                    }
                    catch { }
                }
            }
        }

        public void Send(byte[] buffer, int length)
        {
            m_Offset = 0; //clear RX buffer

            Open();

            int escapedLength = HdlcLayer.PrepareInputBuffer(buffer, length, m_TxBuffer);

            m_SerialPort.Write(m_TxBuffer, 0, escapedLength);
        }

        public int Receive(byte[] buffer)
        {
            Open();

            int lastOffset = 0;

            DateTime startTime = DateTime.UtcNow;

            do
            {
                int byteCount = m_SerialPort.BytesToRead;

                if (byteCount > 0 || m_Offset != lastOffset)
                {
                    m_SerialPort.Read(m_RxBuffer, m_Offset, byteCount);

                    m_Offset += byteCount;

                    if (m_RxBuffer[0] != 0x7E)
                    {
                        throw new InvalidDataException("Bad frame format. No 0x7E mark on start.");
                    }

                    for (int i = 1; i < m_Offset; i++)
                    {
                        if (m_RxBuffer[i] == 0x7E)
                        {
                            int frameLength = HdlcLayer.UnEscapeInputBuffer(m_RxBuffer, i, buffer);

                            int startOfNextFrame = i + 1;
                            
                            Buffer.BlockCopy(m_RxBuffer, startOfNextFrame, m_RxBuffer, 0, m_Offset - startOfNextFrame);

                            m_Offset -= startOfNextFrame;

                            if (HdlcLayer.CheckCrc(buffer, frameLength))
                            {
                                return frameLength;
                            }
                            else
                            {
                                throw new InvalidDataException("Bad Crc of received frame.");
                            }
                        }
                    }

                    lastOffset = m_Offset;
                }

                if (startTime + m_ReadTimeout < DateTime.UtcNow)
                {
                    throw new TimeoutException(string.Format("Receive timeout"));
                }

                Thread.Sleep(50);

            } while (true);
        }
    }
}
